//mod_amf.c is module for Apache httpd web server for detect mobile devices
//Copyright (C) 2009-2010-2011-2012-2013-2014  Idel Fuschini
//
//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU Affero General Public License as
//published by the Free Software Foundation, either version 3 of the
//License, or (at your option) any later version.
//
//Organizations that intend to use this file in it commercially
//or under different licensing terms should contact ApacheMobileFilter
//(http://www.apachemobilefilter.org/contacts/) to learn about licensing options for AMF.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Affero General Public License for more details.
//
//You should have received a copy of the GNU Affero General Public License
//along with this program.  If not, see <http://www.gnu.org/licenses/>.



#include "mod_amf.h"
#include <ctype.h>
#include <fcntl.h>
#include <httpd.h>
#include <http_config.h>
#include <http_log.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>

#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include "apr_env.h"
#include "apr_pools.h"

#define BUFFER_SIZE 1024
#define MAX_SIZE 10000
#define MAX_ERROR_MSG 0x1000
#define AMF_VERSION "1.0.1"

#define AMF_HOST "www.apachemobilefilter.org"

#define ISMOBILE_URL "GET /param/litemobiledetection.config \x0D\x0A\n\x0D\x0A\n"
#define ISTABLET_URL "GET /param/litetabletdetection.config \x0D\x0A\n\x0D\x0A\n"
#define ISTOUCH_URL "GET /param/litetouchdetection.config \x0D\x0A\n\x0D\x0A\n"

int setFullBrowser=0;
int first=0;
char *isMobileString, *isTabletString, *isTouchString;
char KeyFullBrowser[MAX_SIZE];
//static int handlerAMF(request_rec * r);
static int handlerAMF(request_rec* r)
{
    char *IsMobile="false";
    char *IsTablet="false";
    char *IsTouch="false";
	const char *userAgent = NULL;
    const char *x_operamini_phone_ua = NULL;
    const char *x_operamini_ua = NULL;
    const char *x_user_agent = NULL;
    int n=0;
    if (r->headers_in != NULL)
    {
        userAgent = apr_table_get(r->headers_in, "User-Agent");
        //x_user_agent = apr_table_get(r->headers_in, "X-Device-User-Agent");
        x_operamini_phone_ua = apr_table_get(r->headers_in, "X-OperaMini-Phone-Ua");
        x_operamini_ua = apr_table_get(r->headers_in, "X-OperaMini-Ua");
        n++;
    }
    // verify for opera mini browser
    char user_agent[strlen(apr_table_get(r->headers_in,"User-Agent"))];
    strcpy(user_agent,apr_table_get(r->headers_in,"User-Agent"));
   if (x_user_agent) {
        strcpy(user_agent,x_user_agent);
    }
    if (x_operamini_phone_ua) {
        strcpy(user_agent,x_operamini_phone_ua);
    }
    int loop=0;
    char string[strlen(user_agent)];
    for(loop=0;loop<strlen(user_agent);loop++)
        string[loop]=tolower(user_agent[loop]);
    if (checkIsMobile(string)==1) {
        IsMobile="true";
        if (checkIsTablet(string)==1) {
            IsTablet="true";
        }
        if (checkIsTouch(string)==1) {
            IsTouch="true";
        }
    }
    char buffer [1000];
    //void *v;
    //apr_table_t *headers = (apr_table_t *)v;
    //apr_table_t *methods = apr_table_make(r->pool, 12);
    apr_table_t *e = r->subprocess_env;
    apr_table_setn(e, "AMF_ID", "mod_amf_detection");
    //apr_table_setn(r->notes, "AMF_DEVICE_IS_MOBILE", IsMobile);
    //apr_table_setn(r->headers_in, "AMF_DEVICE_IS_MOBILE", IsMobile);

    apr_table_setn(e, "AMF_DEVICE_IS_MOBILE", IsMobile);
    apr_table_setn(e, "AMF_DEVICE_IS_TABLET", IsTablet);
    apr_table_setn(e, "AMF_DEVICE_IS_TOUCH", IsTouch);
    apr_table_setn(e, "SERVER_SOFTWARE", "mod_amf 1.00");
    
    if (setFullBrowser==1) {
        if (r->args) {
            if (checkQueryStringIsFull (r->args)==1) {
                apr_table_add(r->headers_out, "Set-Cookie","amfFull=true; path=/;");
                apr_table_add(e, "AMF_FORCE_TO_DESKTOP", "true");
            }
        }
        if (get_cookie_param(r)==1) {
            apr_table_add(e, "AMF_FORCE_TO_DESKTOP", "true");
        }
    }
    apr_table_setn(e, "AMF_VER", AMF_VERSION);
    apr_table_set(r->headers_out, "AMFplus-Ver", AMF_VERSION);
    sprintf (buffer, "<!-- %d: AMF Vers. %s -->\n",first, AMF_VERSION);
	return DECLINED;
}
/* Compile the regular expression described by "regex_text" into
 "r". */

int compile_regex (regex_t * r, const char * regex_text)
{
    int status = regcomp (r, regex_text, REG_EXTENDED|REG_NEWLINE);
    if (status != 0) {
        char error_message[MAX_ERROR_MSG];
        regerror (status, r, error_message, MAX_ERROR_MSG);
        printf ("Regex error compiling '%s': %s\n",
                regex_text, error_message);
        return 1;
    }
    return 0;
}

/*
 Match the string in "to_match" against the compiled regular
 expression in "r".
 */

int match_regex (regex_t * r, const char * to_match)
{
    /* "P" is a pointer into the string which points to the end of the
     previous match. */
    const char * p = to_match;
    /* "N_matches" is the maximum number of matches allowed. */
    const int n_matches = 10;
    /* "M" contains the matches found. */
    regmatch_t m[n_matches];
	int nomatch = regexec (r, p, n_matches, m, 0);
    if (nomatch) {
        //printf ("No more matches.\n");
        return 1;
    } else {
    	return 0;
    }
    return 0;
}

/* INTERNET CONNECTION */
int socket_connect(char *host, in_port_t port){
    struct hostent *hp;
    struct sockaddr_in addr;
    int on = 1, sock;
    
    if((hp = gethostbyname(host)) == NULL){
        herror("gethostbyname");
        //exit(1);
    } else {
        bcopy(hp->h_addr, &addr.sin_addr, hp->h_length);
        
        addr.sin_port = htons(port);
        addr.sin_family = AF_INET;
        sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
        setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (const char *)&on, sizeof(int));
        if(sock == -1){
            perror("setsockopt");
            exit(1);
        } else {
            if(connect(sock, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)) == -1){
                perror("connect");
                exit(1);
            }
        }
    }
    return sock;
}

int get_cookie_param(request_rec *r)
{
    const char *cookies;
    const char *start_cookie;
    int returnValue=0;
    
    if ((cookies = apr_table_get(r->headers_in, "Cookie"))) {
        char cookiesCompare[strlen(cookies)];
        strcpy(cookiesCompare,cookies);
        if (compare("amfFull=true",cookiesCompare)==1) {
            returnValue=1;
        }
    
    }
    return returnValue;
}

int checkQueryStringIsFull (char *queryString) {
    int returnValue=0;
    if (compare("desktop=true",queryString)==1) {
        returnValue=1;
    }
    return returnValue;
}
int checkIsMobile (char *userAgent) {
    int returnValue=0;
    char * pch;
    char * stringApp=NULL;
    char * copy = malloc(strlen(isMobileString) + 1);
    strcpy(copy, isMobileString);
    pch = strtok (copy,",");
    int count=0;
    while (pch != NULL)
    {
        regex_t r;
        if (compile_regex(& r, pch)==0) {
            if (match_regex(& r, userAgent)==0) {
                return 1;
            }
        }
        regfree (& r);
        pch = strtok (NULL, ",");
    }
    free(copy);
    free(pch);

    return returnValue;
}
int checkIsTouch (char *userAgent) {
    int returnValue=0;
    char * pch;
    char * copy = malloc(strlen(isTouchString) + 1);
    strcpy(copy, isTouchString);
    pch = strtok (copy,",");
    while (pch != NULL)
    {
        regex_t r;
        if (compile_regex(& r, pch)==0) {
            if (match_regex(& r, userAgent)==0) {
                return 1;
            }
        }
        regfree (& r);
        pch = strtok (NULL, ",");
    }
    free(copy);
    free(pch);
    return returnValue;
}

int checkIsTablet (char *userAgent) {
    int returnValue=0;
    char * pch;
    char * copy = malloc(strlen(isTabletString) + 1);
    strcpy(copy, isTabletString);
    pch = strtok (copy,",");
    
    while (pch != NULL)
    {
        regex_t r;
        if (compile_regex(& r, pch)==0) {
            if (match_regex(& r, userAgent)==0) {
                return 1;
            }
        }
        regfree (& r);
        pch = strtok (NULL, ",");
    }
    free(copy);
    free(pch);
    return returnValue;
}
int compare (char *stringToSearch, char *userAgentSearch) {

    int returnValue=0;
	char *stringCompare;
    char *firstChar=substring(stringToSearch, 0,1);
    char *stringCompareFirst=strstr(firstChar,"^");
    if (stringCompareFirst !=NULL) {
        printf("INIZIA: %zd,%zd\n",strlen(stringToSearch) - 1 , strlen(userAgentSearch));
        if (strlen(stringToSearch) - 1 <= strlen(userAgentSearch)) {
            char *stringToSearchSub=substring(stringToSearch, 1,strlen(stringToSearch) - 1);
            char *userAgentCompare=substring(userAgentSearch, 0,strlen(stringToSearchSub));
            char *firstCompareStartWith=strstr(stringToSearchSub,userAgentCompare);
            if (firstCompareStartWith != NULL) {
                returnValue=1;
            }
        }
    } else {
        stringCompare=strstr(userAgentSearch,stringToSearch);
        if(stringCompare != NULL) {
            printf("CONTIENE\n");
            returnValue=1;
        }
    }
    return returnValue;
}

size_t write_data(void *ptr, size_t size, size_t nmemb, FILE *stream) {
    size_t written;
    written = fwrite(ptr, size, nmemb, stream);
    return written;
}
char* substring(const char* str, size_t begin, size_t len)
{
    if (str == 0 || strlen(str) == 0 || strlen(str) < begin || strlen(str) < (begin+len))
        return 0;
    
    return strndup(str + begin, len);
}

/* DIRECTIVE HTTPD.CONF */

static const char *set_fullDesktop(cmd_parms *parms, void *dummy, int flag)
{
    setFullBrowser=flag;
    printf ("AMFFullBrowser is correctly setted\n");
    return NULL;
}

static const char *set_fullBrowserKey(cmd_parms *cmd, void *dummy, const char *map)
{
    sprintf(KeyFullBrowser,"%s",map);
    printf ("AMFKeyFullBrowser is %s \nFor access the device to fullbrowser set the link: <url>%s=true\n",KeyFullBrowser,KeyFullBrowser);
    return NULL;
}
static const char *set_homeDir(cmd_parms *cmd, void *dummy, const char *map)
{
    printf("*************************************************\n");
    printf("*     APACHE MOBILE FILTER + VERSION %s      *\n",AMF_VERSION);
    printf("*************************************************\n");
    
    
    char nameFile[MAX_SIZE];
    FILE *fp;
    char * line = NULL;
    size_t len = 0;
    ssize_t read;
    printf ("AMFHome is correctly setted as: %s\n",map);
    // Detect mobile
    sprintf(nameFile,"%s/litemobiledetection.config",map);
    isMobileString=downloadFile(ISMOBILE_URL,nameFile);
    if (isMobileString != NULL) {
        printf("Configuration for mobile device detection downloaded and saved correctly\n");
    } else {
        printf("Configuration for mobile device detection downloaded failed, try to take old configuration\n");
        fp = fopen(nameFile, "r");
        if (fp == NULL) {
            printf("I couldn't open %s for writing.\n",nameFile );
            exit(1);
        }
        char dummy[MAX_SIZE];
        while((read = getline(&line, &len, fp)) != -1) {
            sprintf(dummy,"%s", line);
            
        }
        size_t size=strlen(dummy) + 1;
        isMobileString=malloc(size);
        strcpy(isMobileString,dummy);
        printf("Configuration for mobile device loaded correctly\n");
    }

    // Detect tablet devices
    sprintf(nameFile,"%s/litetabletdetection.config",map);
    isTabletString=downloadFile(ISTABLET_URL,nameFile);
    if (isTabletString != NULL) {
        printf("Configuration for tablet device detection downloaded and saved correctly\n");
    } else {
        printf("Configuration for tablet device detection downloaded failed, try to take old configuration\n");
        fp = fopen(nameFile, "r");
        if (fp == NULL) {
            printf("I couldn't open %s for writing.\n",nameFile );
            exit(1);
        }
        char dummy[MAX_SIZE];
        while((read = getline(&line, &len, fp)) != -1) {
            sprintf(dummy,"%s", line);
            
        }
        size_t size=strlen(dummy) + 1;
        isTabletString=malloc(size);
        strcpy(isTabletString,dummy);
        printf("Configuration for tablet device loaded correctly\n");
    }

    // Detect touch
    sprintf(nameFile,"%s/litetouchdetection.config",map);
    isTouchString=downloadFile(ISTOUCH_URL,nameFile);
    if (isTouchString != NULL) {
        printf("Configuration for touch  device detection downloaded and saved correctly\n");
    } else {
        printf("Configuration for touch  device detection downloaded failed, try to take old configuration\n");
        fp = fopen(nameFile, "r");
        if (fp == NULL) {
            printf("I couldn't open %s for writing.\n",nameFile );
            exit(1);
        }
        char dummy[MAX_SIZE];
        while((read = getline(&line, &len, fp)) != -1) {
            sprintf(dummy,"%s\n", line);
        }
        size_t size=strlen(dummy) + 1;
        isTouchString=malloc(size);
        strcpy(isTouchString,dummy);
        printf("Configuration for touch device loaded correctly\n");
    }    
    return NULL;
}



char *downloadFile (char *URI, char fileName[]) {
    int fd;
    FILE *fp;
    char buffer[BUFFER_SIZE];
    char *address;
    char dummy[MAX_SIZE]="";
    char *returnValue=NULL;
    fd = socket_connect(AMF_HOST, 80);
    address=URI;
    if (fd == 6) {
        write(fd, address, strlen(address));
        
        bzero(buffer, BUFFER_SIZE);
        while(read(fd, buffer, BUFFER_SIZE - 1) != 0){
            strcat(dummy, buffer);
            bzero(buffer, BUFFER_SIZE);
        }
        shutdown(fd, SHUT_RDWR);
        size_t size=strlen(dummy) + 1;
        returnValue=malloc(size);
        strcpy(returnValue,dummy);
        fp = fopen(fileName, "w");
        if (fp == NULL) {
            exit(1);
        }
        fprintf(fp, "%s",returnValue);
        fclose(fp);
    }
    close(fd);
    return returnValue;
}
static int amf_per_dir(request_rec * r)
{
    
    return handlerAMF(r);
}
static int amf_post_read_request(request_rec * r)
{
    return handlerAMF(r);
}
static void amf_child_init(apr_pool_t * p, server_rec * s)
{

}

static int amf_post_config(apr_pool_t * p, apr_pool_t * plog,
                  apr_pool_t * ptemp, server_rec * s)
{
    return OK;
}
/*static void amf_server_init(apr_pool_t * p, server_rec * s)
{

} */

///////////////////////
static void register_hooks(apr_pool_t *p)
{
    /* make sure we run before mod_rewrite's handler */
    static const char * const aszSucc[] = {"mod_setenvif.c", "mod_rewrite.c", NULL };
    //ap_hook_handler(handlerAMF, NULL, NULL, APR_HOOK_MIDDLE - 1);
    ap_hook_header_parser(amf_per_dir, NULL, aszSucc, APR_HOOK_FIRST);
    /* and the servectly wide hook, after reading the request. Perfecly
     * suitable to serve serverwide mod_rewrite actions
     */
    ap_hook_post_read_request(amf_post_read_request, NULL, aszSucc, APR_HOOK_MIDDLE);
}


static const command_rec amf_cmds [] = {
    AP_INIT_TAKE1("AMFHome", set_homeDir, NULL,
                 RSRC_CONF, "Define HomeDirectory"),
    AP_INIT_FLAG("AMFFullBrowser", set_fullDesktop, NULL,
                 RSRC_CONF, "Define if you want to manage fullbrowser"),
    AP_INIT_TAKE1("AMFFullBrowserAccessKey", set_fullBrowserKey, NULL,
                  RSRC_CONF, "Define the accesskey for your fullbrowser"),
    {NULL}
};

module AP_MODULE_DECLARE_DATA amf_module = 
{ 
	STANDARD20_MODULE_STUFF,
    NULL, /* per-directory config creator */
    NULL,     /* dir config merger */
    NULL,     /* server config creator */
    NULL,     /* server config merger */
    amf_cmds, /* command table */
    register_hooks
};
